/*------------------------------------------------------------------------
 |  FILE            adjclock.c
 |
 |  DESCRIPTION     Adjusts the computer's clock. The file ADJCLOCK.TXT
 |                  is taken as documentation in addition to the few
 |                  comments at the start of each function.
 |
 |  WRITTEN BY      Sverre H. Huseby <shh@thathost.com>
 +----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <io.h>
#include <fcntl.h>
#include <dir.h>

#define PROG_NAME  "ADJCLOCK"
#define PROG_VER   "v2.0"
#define PROG_DATE  "2000-09-11"
//#define BETA

#define DATE_PARAM 0x01
#define TIME_PARAM 0x02

/*-----------------------------------------------------------------------+
|  PRIVATE DATA                                                          |
+-----------------------------------------------------------------------*/

static char datFileName[MAXPATH];
static int  datFile, fileCreated = 0;
static struct {
    char   id[31];
    time_t mStart,       /* Measure start */
           mEnd,         /* Measure end (incorrect time) */
           mCorrect,     /* Measure end (correct time) */
           last;         /* Last time adjusted */
} FileCont = { PROG_NAME " data file (SHH)", 0, 0, 0, 0 };
static time_t currTime;
static int minInaccuracy = 5;  /* Need at least this numer of seconds
                                * off to make an adjustment */


/*-----------------------------------------------------------------------+
|  PRIVATE FUNCTIONS                                                     |
+-----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
 |
 |  NAME          fatal
 |
 |  FUNCTION      Show given message on stderr and abort the program.
 |
 |  INPUT         format, ...  Text formatted as for printf()
 |
 |  RETURNS       Never returns
 |
 */
static void
fatal(const char *format, ...)
{
    va_list ap;

    fprintf(stderr, "\n%s: ", PROG_NAME);
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    fprintf(stderr, "\n");
    exit(1);
}

/*-------------------------------------------------------------------------
 |
 |  NAME          makeTime
 |
 |  FUNCTION      Convert dmy hm to a time() -compatible value.
 |
 |  INPUT         day     1-31
 |                month   1-12
 |                year    0-> (relative to 1900)
 |                hour    0-23
 |                min     0-59
 |                sec     0-59
 |
 |                Values of -1 indicate: Use todays value.
 |
 |  RETURNS       time_t value.
 |
 */
static time_t
makeTime(int day, int month, int year, int hour, int min, int sec)
{
    struct tm *t;


    t = localtime(&currTime);

    if (day != -1)
        t->tm_mday = day;
    if (month != -1)
        t->tm_mon = month - 1;
    if (year != -1)
        t->tm_year = year;
    if (hour != -1)
        t->tm_hour = hour;
    if (min != -1)
        t->tm_min = min;
    if (sec != -1)
        t->tm_sec = sec;

    return mktime(t);
}

/*-------------------------------------------------------------------------
 |
 |  NAME          strTime
 |
 |  FUNCTION      Return (static) string representation of the
 |                given time. (Date and Time).
 |
 |  INPUT         t       Time to convert to ascii.
 |
 |  RETURNS       String form of given time.
 |
 */
static char *
strTime(clock_t t)
{
    static char ret[81];
    struct tm *tbuf;
    int    day, month, year, hour, min, sec;


    tbuf  = localtime(&t);
    day   = tbuf->tm_mday;
    month = tbuf->tm_mon + 1;
    year  = 1900 + tbuf->tm_year;

    hour  = tbuf->tm_hour;
    min   = tbuf->tm_min;
    sec   = tbuf->tm_sec;

    sprintf(ret, "%04d-%02d-%02d %02d:%02d:%02d",
            year, month, day, hour, min, sec);

    return ret;
}

/*-------------------------------------------------------------------------
 |
 |  NAME          setClock
 |
 |  FUNCTION      Set new computer clock time.
 |
 |  INPUT         t       New time.
 |                quiet   Supress messages?
 |
 */
static void
setClock(time_t t, int quiet)
{
    if (!quiet)
        printf("  Original clock: %s\n", strTime(currTime));
    stime(&t);
    currTime = t;
    if (!quiet)
        printf("  Adjusted to:    %s\n", strTime(currTime));
}

/*-------------------------------------------------------------------------
 |
 |  NAME          openDat
 |
 |  FUNCTION      Open and possibly read the contents of the DAT-file.
 |                the file is left open with the file pointer at the
 |                start. The file is opened in read/write mode.
 |
 |  INPUT         mustexist
 |                        Must the file exist? (Abort if not)
 |
 */
static void
openDat(int mustexist)
{
    /* Possibly check existence of file. */
    if (mustexist && access(datFileName, 0) != 0)
        fatal("No file %s found. Use `" PROG_NAME " n' to create.",
              datFileName);

    /* Open and possibly read file. */
    if ((datFile = _open(datFileName, O_RDWR)) != -1) {
        if (_read(datFile, &FileCont, sizeof(FileCont)) < sizeof(FileCont))
            fatal("Error reading %s", datFileName);
        lseek(datFile, 0L, SEEK_SET);
    } else if ((datFile = _creat(datFileName, 0)) == -1)
        fatal("Unable to open/create %s", datFileName);
    else
        fileCreated = 1;
}

/*-------------------------------------------------------------------------
 |
 |  NAME          closeDat
 |
 |  FUNCTION      Write DAT-file, and close it.
 |
 */
static void
closeDat(void)
{
    /* Write and close file. */
    if (_write(datFile, &FileCont, sizeof(FileCont)) < sizeof(FileCont))
        fatal("Error writing %s", datFileName);
    _close(datFile);
}

/*-------------------------------------------------------------------------
 |
 |  NAME          isDate
 |
 |  FUNCTION      Check if a string is a date, and if it is, get
 |                its elements. There must be no white space characters
 |                in the string.
 |
 |  INPUT         s       String to check
 |                d       Pointer to location for day.
 |                m       Pointer to location for month.
 |                y       Pointer to location for year.
 |
 |  RETURNS       0   No date. dmy not changed.
 |                1   Date.
 |
 */
static int
isDate(char *s, int *d, int *m, int *y)
{
    long dd = 0L, mm = 0L, yy = 0L;


    if (!isdigit(*s))
        return 0;

    while (isdigit(*s))
        yy = 10L * yy + *s++ - '0';
    if (*s != '-')
        return 0;

    if (!isdigit(*++s))
        return 0;
    while (isdigit(*s))
        mm = 10L * mm + *s++ - '0';
    if (*s != '-')
        return 0;

    if (!isdigit(*++s))
        return 0;
    while (isdigit(*s))
        dd = 10L * dd + *s++ - '0';
    if (*s)
        return 0;

    if (yy >= 1900L)
        yy -= 1900L;
    else if (yy >= 0L && yy <= 37L)
        yy += 100L;
    if (dd < 1L || dd > 31L || mm < 1L || mm > 12L || yy < 70L || yy > 137L)
        fatal("Illegal date");

    *d = (int) dd;
    *m = (int) mm;
    *y = (int) yy;

    return 1;
}

/*-------------------------------------------------------------------------
 |
 |  NAME          isTime
 |
 |  FUNCTION      Check if a string is a time, and if it is, get
 |                its elements. There must be no white space characters
 |                in the string.
 |
 |  INPUT         s       String to check
 |                h       Pointer to location for hours.
 |                m       Pointer to location for minutes.
 |                c       Pointer to location for seconds.
 |
 |  RETURNS       0   No time. hm not changed.
 |                1   Time.
 |
 */
static int
isTime(char *s, int *h, int *m, int *c)
{
    long hh = 0L, mm = 0L, ss = 0L;


    if (!isdigit(*s))
        return 0;

    while (isdigit(*s))
        hh = 10L * hh + *s++ - '0';
    if (*s != ':')
        return 0;

    if (!isdigit(*++s))
        return 0;
    while (isdigit(*s))
        mm = 10L * mm + *s++ - '0';
    if (*s) {
        if (*s != ':')
            return 0;
        if (!isdigit(*++s))
            return 0;
        while (isdigit(*s))
            ss = 10L * ss + *s++ - '0';
        if (*s)
            return 0;
    }

    if (hh < 0L || hh > 23L || mm < 0L || mm > 59L || ss < 0L || ss > 59L)
        fatal("Illegal time");

    *h = (int) hh;
    *m = (int) mm;
    *c = (int) ss;

    return 1;
}

/*-------------------------------------------------------------------------
 |
 |  NAME          getDateTime
 |
 |  FUNCTION      Search list of arguments for date or time. Convert
 |                to time_t -value. If eiter (or both) is not given,
 |                todays time is used for the missing values.
 |
 |                !!! Matching arguments are removed from the list !!!
 |
 |  INPUT         parc    Pointer to number of parameters.
 |                parv    Parameters to this command.
 |                t       Pointer to location for returned time.
 |
 |  RETURNS       Bit-combination of DATE_PARAM and TIME_PARAM to tell
 |                what was found.
 |
 */
static int
getDateTime(int *parc, char *parv[], time_t *t)
{
    int q, w, ret = 0, flag,
        day = -1, month = -1, year = -1, hour = -1, min = -1, sec = -1;

    for (q = 0; q < *parc; q++) {
        flag = 0;
        if (isDate(parv[q], &day, &month, &year)) {
            if (ret & DATE_PARAM)
                fatal("Only one date can be given");
            flag = ret |= DATE_PARAM;
        } else if (isTime(parv[q], &hour, &min, &sec)) {
            if (ret & TIME_PARAM)
                fatal("Only one time can be given");
            flag = ret |= TIME_PARAM;
        }
        if (flag) {
            for (w = q; w < *parc - 1; w++)
                parv[w] = parv[w + 1];
            --*parc;
            --q;
        }
    }

    *t = makeTime(day, month, year, hour, min, sec);

    return ret;
}

/*-------------------------------------------------------------------------
 |
 |  NAME          abortIfOptions
 |
 |  FUNCTION      Abort if more options left in option list. The
 |                program was started with too many/unknown options.
 |
 |  INPUT         parc    Number of parameters to this command.
 |                parv    Parameters to this command.
 |
 */
static void
abortIfOptions(int parc, char *parv[])
{
    int q;
    char s[201] = "Illegal option";


    if (!parc)
        return;

    strcat(s, parc == 1 ? ":" : "s:");
    for (q = 0; q < parc; q++) {
        strcat(s, " ");
        strcat(s, parv[q]);
    }

    fatal("%s", s);
}

/*-------------------------------------------------------------------------
 |
 |  NAME          adjust
 |
 |  FUNCTION      Perform clock adjusting.
 |
 |  INPUT         parc    Number of parameters to this command.
 |                parv    Parameters to this command.
 |
 */
static void
adjust(int parc, char *parv[])
{
    int    q, w, quiet = 0, flag;
    char   *s;
    double compdiff, realdiff, mcomp, mreal;
    long   diff;
    time_t newtime, t;


    flag = getDateTime(&parc, parv, &t);
    if (flag)
        fatal("No date or time should be given to the a-command");
    for (q = 0; q < parc; q++) {
        if (stricmp(parv[q], "/Q") == 0 || stricmp(parv[q], "-Q") == 0) {
            for (w = q; w < parc - 1; w++)
                parv[w] = parv[w + 1];
            --parc;
            --q;
            quiet = 1;
        } else if (isdigit(parv[q][0])) {
            s = parv[q];
            minInaccuracy = 0;
            while (isdigit(*s))
                minInaccuracy = 10L * minInaccuracy + *s++ - '0';
            if (*s != '\0' || minInaccuracy < 1)
                fatal("Illegal minimum inaccuracy");
            for (w = q; w < parc - 1; w++)
                parv[w] = parv[w + 1];
            --parc;
            --q;
        }
    }
    abortIfOptions(parc, parv);
    openDat(1);
    if (!FileCont.mEnd)
        fatal("Measure currently running. Unable to adjust yet");

    mcomp    = (double) FileCont.mEnd - (double) FileCont.mStart;
    mreal    = (double) FileCont.mCorrect - (double) FileCont.mStart;
    compdiff = (double) currTime - (double) FileCont.last;
    realdiff = (mreal / mcomp) * compdiff;
    newtime  = FileCont.last + (long) floor(realdiff + 0.5);
    diff     = realdiff - compdiff;

    if (!quiet)
        printf("\n  Adjusting computer clock:\n");
    if (labs(diff) < (long) minInaccuracy) {
        if (!quiet)
            printf("  Not enough to adjust yet.\n"
                   "  Waiting for at least %d seconds inaccuracy.\n\n",
                   minInaccuracy);
    } else {
        setClock(newtime, quiet);
        if (!quiet)
            printf("  (%s %ld seconds)\n\n",
                   diff < 0L ? "Subtracted" : "Added",
                   labs(diff));
        FileCont.last = newtime;
    }
    closeDat();
}

/*-------------------------------------------------------------------------
 |
 |  NAME          newMeasure
 |
 |  FUNCTION      Start new measuring.
 |
 |  INPUT         parc    Number of parameters to this command.
 |                parv    Parameters to this command.
 |
 */
static void
newMeasure(int parc, char *parv[])
{
    int    flag;
    time_t t;


    flag = getDateTime(&parc, parv, &t);
    abortIfOptions(parc, parv);
    openDat(0);
    printf("\n");
    if (flag)
        setClock(t, 0);
    printf("  Starting %s inaccuracy measure at %s.\n",
           fileCreated ? "first" : "new",
           strTime(currTime));
    if (fileCreated)
        printf("  Created file %s for data.\n", datFileName);
    FileCont.mStart = currTime;
    FileCont.mEnd = FileCont.mCorrect = FileCont.last = 0;
    printf("\n");
    closeDat();
}

/*-------------------------------------------------------------------------
 |
 |  NAME          set
 |
 |  FUNCTION      Set date/time.
 |
 |  INPUT         parc    Number of parameters to this command.
 |                parv    Parameters to this command.
 |
 */
static void
set(int parc, char *parv[])
{
    int    flag;
    long   diff;
    double mcomp, mreal;
    time_t t;


    flag = getDateTime(&parc, parv, &t);
    abortIfOptions(parc, parv);
    openDat(1);
    if (!flag)
        fatal("Specify new date and/or time");
    printf("\n");
    if (!FileCont.mEnd) {
        diff = currTime - t;
        if (currTime < FileCont.mStart)
            printf(
              "\n  Aborted: Something is wrong. The current time seems to be\n"
              "  _before_ the measure started! Please start a new measure.\n"
            );
        else if (t < FileCont.mStart)
            printf(
              "\n  Aborted: The given time is _before_ the measure started!\n"
              "  Please try again.\n"
            );
        else if (currTime - FileCont.mStart < 3600)
            printf(
              "\n  Aborted: Please be patient. You just started this.\n"
              "  Wait some more!\n"
            );
        else if (labs(diff) < (long) minInaccuracy)
            printf(
              "\n  Aborted: Need at least %d seconds inaccuracy.\n"
              "  Wait some more.\n",
              minInaccuracy
            );
        else {
            printf("  Ending inaccuracy measure started %s\n",
                   strTime(FileCont.mStart));
            FileCont.mEnd = currTime;
            FileCont.mCorrect = FileCont.last = t;
            setClock(t, 0);
            mcomp = (double) FileCont.mEnd - (double) FileCont.mStart;
            mreal = (double) FileCont.mCorrect - (double) FileCont.mStart;
            printf("  Inaccuracy:     %+G seconds pr day.\n",
                   ((mcomp - mreal) / mreal) * 24.0 * 3600.0);
        }
    } else {
        setClock(t, 0);
        FileCont.last = t;
    }
    printf("\n");
    closeDat();
}

/*-------------------------------------------------------------------------
 |
 |  NAME          update
 |
 |  FUNCTION      Update date/time.
 |
 |  INPUT         parc    Number of parameters to this command.
 |                parv    Parameters to this command.
 |
 */
static void
update(int parc, char *parv[])
{
    int    flag;
    time_t t;


    flag = getDateTime(&parc, parv, &t);
    if (flag)
        fatal("No date or time should be given to the u-command");
    abortIfOptions(parc, parv);
    openDat(1);
    if (!FileCont.mEnd)
        fatal("Measure currently running. You should _not_ change "
              "the clock now!");
    printf("\n  Updated correct time to %s\n\n", strTime(currTime));
    FileCont.last = currTime;
    closeDat();
}

/*-------------------------------------------------------------------------
 |
 |  NAME          usage
 |
 |  FUNCTION      Display usage and exit.
 |
 |  RETURNS       Never returns
 |
 */
static void
usage(void)
{
    printf(
        PROG_NAME " " PROG_VER "  --  (C) " PROG_DATE " - "
                               "Sverre H. Huseby, Norway.\n"
      #ifdef BETA
"\n"
"    This is a BETA-version.  See the file BETA.TXT for details.\n"
"    Should not be used longer than 3 months after the above date.\n"
      #endif
"\n"
"Program for automatic (non-resident) PC-clock adjustment.  When called\n"
"periodically, the program tries to compensate for clock inaccuracy.\n"
"\n"
"Usage: " PROG_NAME " <command> [parameters]\n"
"\n"
"  Valid commands and parameters:\n"
"    a [mininacc] [-q]   Adjust clock.  Disable messages with -q.\n"
"    n [date] [time]     Start New measure.\n"
"    s [date] [time]     Set date/time.\n"
"    u                   Update date and time.\n"
"\n"
"Date should be given as YYYY-MM-DD and time as HH:MM or HH:MM:SS where\n"
"HH is 0-23.  Mininacc is the minimum inaccuracy needed for adjustment,\n"
"which defaults to %d seconds.  See the file " PROG_NAME ".TXT for details.\n"
"\n",
        minInaccuracy
    );

    exit(1);
}



/*-----------------------------------------------------------------------+
|  PUBLIC FUNCTIONS                                                      |
+-----------------------------------------------------------------------*/

int
main(int argc, char *argv[])
{
    char drv[MAXDRIVE], dir[MAXDIR];


    if (argc == 1)
        usage();

    /* Get location and name of DAT-file. */
    if (_osmajor < 3)
        fatal("Need MSDOS version 3.0 or higher");
    fnsplit(argv[0], drv, dir, (char *) NULL, (char *) NULL);
    fnmerge(datFileName, drv, dir, PROG_NAME, ".DAT");

    /* Get current time. */
    currTime = time((time_t *) NULL);

    /* Get and perform command. */
    if (stricmp(argv[1], "a") == 0)
        adjust(argc - 2, argv + 2);
    else if (stricmp(argv[1], "n") == 0)
        newMeasure(argc - 2, argv + 2);
    else if (stricmp(argv[1], "s") == 0)
        set(argc - 2, argv + 2);
    else if (stricmp(argv[1], "u") == 0)
        update(argc - 2, argv + 2);
    else
        fatal("Unknown command %s", argv[1]);

    return 0;
}
